from __future__ import annotations

from typing import Any

from pydantic import BaseModel, ConfigDict, Field, RootModel
from typing_extensions import Annotated


class DescriptionLeaf(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[str, Field(title="DescriptionLeaf")]
    """
    A textual description of the interface.

    A server implementation MAY map this leaf to the ifAlias
    MIB object.  Such an implementation needs to use some
    mechanism to handle the differences in size and characters
    allowed between this leaf and ifAlias.  The definition of
    such a mechanism is outside the scope of this document.

    Since ifAlias is defined to be stored in non-volatile
    storage, the MIB implementation MUST map ifAlias to the
    value of 'description' in the persistently stored
    datastore.

    Specifically, if the device supports ':startup', when
    ifAlias is read the device MUST return the value of
    'description' in the 'startup' datastore, and when it is
    written, it MUST be written to the 'running' and 'startup'
    datastores.  Note that it is up to the implementation to

    decide whether to modify this single leaf in 'startup' or
    perform an implicit copy-config from 'running' to
    'startup'.

    If the device does not support ':startup', ifAlias MUST
    be mapped to the 'description' leaf in the 'running'
    datastore.
    """


class EnabledLeaf(RootModel[bool]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[bool, Field(title="EnabledLeaf")]
    """
    This leaf contains the configured, desired state of the
    interface.

    Systems that implement the IF-MIB use the value of this
    leaf in the 'running' datastore to set
    IF-MIB.ifAdminStatus to 'up' or 'down' after an ifEntry
    has been initialized, as described in RFC 2863.

    Changes in this leaf in the 'running' datastore are
    reflected in ifAdminStatus, but if ifAdminStatus is
    changed over SNMP, this leaf is not affected.
    """


class LoopbackModeLeaf(RootModel[bool]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[bool, Field(title="Loopback-modeLeaf")]
    """
    When set to true, the interface is logically looped back,
    such that packets that are forwarded via the interface
    are received on the same interface.
    """


class MtuLeaf(RootModel[int]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[int, Field(ge=0, le=65535, title="MtuLeaf")]
    """
    Set the max transmission unit size in octets
    for the physical interface.  If this is not set, the mtu is
    set to the operational default -- e.g., 1514 bytes on an
    Ethernet interface.
    """


class NameLeaf(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[str, Field(title="NameLeaf")]
    """
    The name of the interface.

    A device MAY restrict the allowed values for this leaf,
    possibly depending on the type of the interface.
    For system-controlled interfaces, this leaf is the
    device-specific name of the interface.  The 'config false'
    list interfaces/interface[name]/state contains the currently
    existing interfaces on the device.

    If a client tries to create configuration for a
    system-controlled interface that is not present in the
    corresponding state list, the server MAY reject
    the request if the implementation does not support
    pre-provisioning of interfaces or if the name refers to
    an interface that can never exist in the system.  A
    NETCONF server MUST reply with an rpc-error with the
    error-tag 'invalid-value' in this case.

    The IETF model in RFC 7223 provides YANG features for the
    following (i.e., pre-provisioning and arbitrary-names),
    however they are omitted here:

     If the device supports pre-provisioning of interface
     configuration, the 'pre-provisioning' feature is
     advertised.

     If the device allows arbitrarily named user-controlled
     interfaces, the 'arbitrary-names' feature is advertised.

    When a configured user-controlled interface is created by
    the system, it is instantiated with the same name in the
    /interfaces/interface[name]/state list.
    """


class TypeLeaf(RootModel[Any]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[Any, Field(title="TypeLeaf")]
    """
    The type of the interface.

    When an interface entry is created, a server MAY
    initialize the type leaf with a valid value, e.g., if it
    is possible to derive the type from the name of the
    interface.

    If a client tries to set the type of an interface to a
    value that can never be used by the system, e.g., if the
    type is not supported or if the type does not match the
    name of the interface, the server MUST reject the request.
    A NETCONF server MUST reply with an rpc-error with the
    error-tag 'invalid-value' in this case.
    """


class ConfigContainer(BaseModel):
    """
    Configurable items at the global, physical interface
    level
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    name: Annotated[NameLeaf, Field(None, alias="openconfig-interfaces:name")]
    type: Annotated[TypeLeaf, Field(alias="openconfig-interfaces:type")]
    mtu: Annotated[MtuLeaf, Field(None, alias="openconfig-interfaces:mtu")]
    loopback_mode: Annotated[
        LoopbackModeLeaf, Field(False, alias="openconfig-interfaces:loopback-mode")
    ]
    description: Annotated[
        DescriptionLeaf, Field(None, alias="openconfig-interfaces:description")
    ]
    enabled: Annotated[EnabledLeaf, Field(True, alias="openconfig-interfaces:enabled")]


class Model(BaseModel):
    """
    Initialize an instance of this class and serialize it to JSON; this results in a RESTCONF payload.

    ## Tips
    Initialization:
    - all values have to be set via keyword arguments
    - if a class contains only a `root` field, it can be initialized as follows:
        - `member=MyNode(root=<value>)`
        - `member=<value>`

    Serialziation:
    - `exclude_defaults=True` omits fields set to their default value (recommended)
    - `by_alias=True` ensures qualified names are used (necessary)
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    config: Annotated[
        ConfigContainer, Field(None, alias="openconfig-interfaces:config")
    ]


if __name__ == "__main__":
    model = Model(  # type: ignore[call-arg]
        # <Initialize model here>
    )

    restconf_payload = model.model_dump_json(
        exclude_defaults=True, by_alias=True, indent=2
    )

    print(f"Generated output: {restconf_payload}")

    # Send config to network device:
    # from pydantify.utility import restconf_patch_request
    # restconf_patch_request(url='...', user_pw_auth=('usr', 'pw'), data=restconf_payload)



translation_map = {
    # https://www.w3.org/TR/2004/REC-xmlschema-2-20041028/#nt-charProp
    r"\p{L}": r"\w",  # All Letters
    r"\P{L}": r"\W",  # All Letters
    r"\p{Lu}": r"[A-Z]",  # uppercase
    r"\P{Lu}": r"[^A-Z]",  # uppercase
    r"\p{Ll}": r"[a-z]",  # uppercase
    r"\P{Ll}": r"[^a-z]",  # uppercase
    r"\p{N}": r"\d",  # All Numbers
    r"\P{N}": r"\D",  # All Numbers
    r"\p{Nd}": r"\d",  # decimal digit
    r"\P{Nd}": r"\D",  # decimal digit
}


def convert_pattern(pattern: str) -> str:
    """
    Convert xmlschema pattern into python regex.
    ToDo: This is just a quick fix and need to be done better
    """
    for search, replace in translation_map.items():
        pattern = pattern.replace(search, replace)
    return pattern



import os
import pytest

if os.getenv("_PYTEST_RAISE", "0") != "0":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value



from __future__ import annotations

from typing import List

from pydantic import BaseModel, ConfigDict, Field, RootModel
from typing_extensions import Annotated


class DottedQuadType(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
        regex_engine="python-re",
    )
    root: Annotated[
        str,
        Field(
            pattern="^(?=^(([0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])\\.){3}([0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])$).*$"
        ),
    ]
    """
    Four octets written as decimal numbers and separated with the '.' (full stop) character.
    """


class EnabledLeaf(RootModel[bool]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[bool, Field(title="EnabledLeaf")]
    """
    Enable or disable the interface. Example value: true
    """


class NameLeaf(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[str, Field(title="NameLeaf")]
    """
    Interface name. Example value: GigabitEthernet 0/0/0
    """


class SubnetMaskLeaf(RootModel[DottedQuadType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[DottedQuadType, Field(title="Subnet-maskLeaf")]
    """
    Interface subnet mask. Example value: 255.255.255.0
    """


class AddressLeaf(RootModel[DottedQuadType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[DottedQuadType, Field(title="AddressLeaf")]
    """
    Interface IP address. Example value: 10.10.10.1
    """


class InterfaceListEntry(BaseModel):
    """
    Regular IPv4 address with subnet
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    name: Annotated[NameLeaf, Field(alias="interfaces:name")]
    address: Annotated[AddressLeaf, Field(alias="interfaces:address")]
    subnet_mask: Annotated[SubnetMaskLeaf, Field(alias="interfaces:subnet-mask")]
    enabled: Annotated[EnabledLeaf, Field(False, alias="interfaces:enabled")]


class InterfacesContainer(BaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    interface: Annotated[List[InterfaceListEntry], Field(alias="interfaces:interface")]


class Model(BaseModel):
    """
    Initialize an instance of this class and serialize it to JSON; this results in a RESTCONF payload.

    ## Tips
    Initialization:
    - all values have to be set via keyword arguments
    - if a class contains only a `root` field, it can be initialized as follows:
        - `member=MyNode(root=<value>)`
        - `member=<value>`

    Serialziation:
    - `exclude_defaults=True` omits fields set to their default value (recommended)
    - `by_alias=True` ensures qualified names are used (necessary)
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    interfaces: Annotated[
        InterfacesContainer, Field(None, alias="interfaces:interfaces")
    ]


if __name__ == "__main__":
    model = Model(  # type: ignore[call-arg]
        # <Initialize model here>
    )

    restconf_payload = model.model_dump_json(
        exclude_defaults=True, by_alias=True, indent=2
    )

    print(f"Generated output: {restconf_payload}")

    # Send config to network device:
    # from pydantify.utility import restconf_patch_request
    # restconf_patch_request(url='...', user_pw_auth=('usr', 'pw'), data=restconf_payload)



from __future__ import annotations

from enum import Enum
from typing import List

from pydantic import BaseModel, ConfigDict, Field, RootModel
from typing_extensions import Annotated


class NameLeaf(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[str, Field(title="NameLeaf")]
    """
    Interface name
    """


class EnumerationEnum(Enum):
    enable = "enable"
    disable = "disable"


class AdminStateType(RootModel[EnumerationEnum]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: EnumerationEnum
    """
    general admin-state option.
    """


class AdminStateLeaf(RootModel[AdminStateType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[AdminStateType, Field(title="Admin-stateLeaf")]
    """
    The configured, desired state of the interface
    """


class InterfacesListEntry(BaseModel):
    """
    List of configured device interfaces
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    name: Annotated[NameLeaf, Field(None, alias="interfaces:name")]
    admin_state: Annotated[
        AdminStateLeaf, Field(AdminStateLeaf("enable"), alias="interfaces:admin-state")
    ]


class Model(BaseModel):
    """
    Initialize an instance of this class and serialize it to JSON; this results in a RESTCONF payload.

    ## Tips
    Initialization:
    - all values have to be set via keyword arguments
    - if a class contains only a `root` field, it can be initialized as follows:
        - `member=MyNode(root=<value>)`
        - `member=<value>`

    Serialziation:
    - `exclude_defaults=True` omits fields set to their default value (recommended)
    - `by_alias=True` ensures qualified names are used (necessary)
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    interfaces: Annotated[
        List[InterfacesListEntry], Field(alias="interfaces:interfaces")
    ]


if __name__ == "__main__":
    model = Model(
        # <Initialize model here>
    )

    restconf_payload = model.model_dump_json(
        exclude_defaults=True, by_alias=True, indent=2
    )

    print(f"Generated output: {restconf_payload}")

    # Send config to network device:
    # from pydantify.utility import restconf_patch_request
    # restconf_patch_request(url='...', user_pw_auth=('usr', 'pw'), data=restconf_payload)



from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field, RootModel
from typing_extensions import Annotated


class AddressLeaf(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[str, Field(title="AddressLeaf")]
    """
    Target IP address
    """


class PortLeaf(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[str, Field(title="PortLeaf")]
    """
    Target port number
    """


class DestinationContainer(BaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    address: Annotated[AddressLeaf, Field(None, alias="interfaces:address")]
    port: Annotated[PortLeaf, Field(None, alias="interfaces:port")]


class PeerContainer(BaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    destination: Annotated[
        DestinationContainer, Field(None, alias="interfaces:destination")
    ]


class Model(BaseModel):
    """
    Initialize an instance of this class and serialize it to JSON; this results in a RESTCONF payload.

    ## Tips
    Initialization:
    - all values have to be set via keyword arguments
    - if a class contains only a `root` field, it can be initialized as follows:
        - `member=MyNode(root=<value>)`
        - `member=<value>`

    Serialziation:
    - `exclude_defaults=True` omits fields set to their default value (recommended)
    - `by_alias=True` ensures qualified names are used (necessary)
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    peer: Annotated[PeerContainer, Field(None, alias="interfaces:peer")]


if __name__ == "__main__":
    model = Model(  # type: ignore[call-arg]
        # <Initialize model here>
    )

    restconf_payload = model.model_dump_json(
        exclude_defaults=True, by_alias=True, indent=2
    )

    print(f"Generated output: {restconf_payload}")

    # Send config to network device:
    # from pydantify.utility import restconf_patch_request
    # restconf_patch_request(url='...', user_pw_auth=('usr', 'pw'), data=restconf_payload)



import ast
import logging
import os
import sys
from pathlib import Path
from typing import Any, List
from unittest.mock import patch

import pytest
from pydantic import validate_call
from pytest import param
from typing_extensions import Self

LOGGER = logging.getLogger(__name__)


class ParsedAST:
    def __init__(self, file: Path) -> None:
        self.file: Path = file
        with file.open() as f:
            self.ast = ast.parse(f.read())
        self.body: list[ast.stmt] = self.ast.body
        self.classes: dict[str, ast.ClassDef] = dict()
        for c in self.body:
            if isinstance(c, ast.ClassDef):
                try:
                    bases = ", ".join([b.id for b in c.bases])  # type: ignore[attr-defined]
                except AttributeError:
                    bases = ", ".join([b.value.id for b in c.bases])  # type: ignore[attr-defined]
                self.classes[f"{c.name}({bases})"] = c

    @staticmethod
    def get_annotation(annotation: Any) -> ast.expr:
        return annotation.value if isinstance(annotation, ast.Subscript) else annotation

    @staticmethod
    @validate_call
    def assert_python_sources_equal(generated: Path, expected: Path):
        LOGGER.info(f"Output path: {generated}")
        ast1 = ParsedAST(generated)
        ast2 = ParsedAST(expected)
        LOGGER.info(
            f'"Comparing:\n{"Expected":9}: {ast2.classes.keys()}\n{"Got":9}: {ast1.classes.keys()}'
        )
        for c1, c2 in zip(ast1.classes.keys(), ast2.classes.keys()):
            assert (
                c1 == c2
            ), f'Mismatch {c1} vs {c2}\nGot: "{ast1.classes}"\nExpected: "{ast2.classes}"'
        for a, b in zip(ast1.classes.values(), ast2.classes.values()):
            # Compare classes
            assert len(a.body) == len(b.body)
            for a2, b2 in zip(a.body, b.body):
                # Compare class members
                annotation_a = getattr(a2, "annotation", None)
                annotation_b = getattr(b2, "annotation", None)
                # Compare annotation
                assert (annotation_a is None) == (annotation_b is None)
                if annotation_a is not None:
                    # Compare annotated type
                    assert getattr(
                        ast1.get_annotation(annotation_a), "id", None
                    ) == getattr(ast2.get_annotation(annotation_b), "id", None)
        assert len(ast1.body) == len(ast2.body)


def run_pydantify(input_file: Path, output_folder: Path, args: List[str] = []):
    args = [
        sys.argv[0],
        *args,
        f"-i={input_file.parent}",
        f"-o={output_folder}",
        str(input_file),
    ]
    with patch.object(sys, "argv", args):
        from pydantify.main import main

        try:
            main()
        except SystemExit as e:
            assert e.code == 0, f"Pyang exited with errors:\n{e}"


@pytest.fixture(autouse=True)
def reset_optparse():
    from pyang import plugin
    from pydantify.models.base import Node

    # Reset plugins. Otherwise pyang creates cross-test side-effects. TODO: Better way?
    plugin.plugins = []
    Node._name_count = dict()


@pytest.mark.parametrize(
    ("input_dir", "expected_file", "args"),
    [
        param(
            "examples/minimal/interfaces.yang",
            "examples/minimal/expected.py",
            [],
            id="minimal",
        ),
        param(
            "examples/minimal/interfaces.yang",
            "examples/minimal/expected_trimmed.py",
            ["-t=/interfaces/interfaces/address"],
            id="minimal_trimmed",
        ),
        param(
            "examples/minimal/interfaces.yang",
            "examples/minimal/expected_standalone.py",
            ["--standalone"],
            id="minimal_standalone",
        ),
        param(
            "examples/minimal/interfaces.yang",
            "examples/minimal/expected_trimmed.py",
            ["-t=interfaces/interfaces/address"],
            id="minimal_trimmed without leading /",
        ),
        param(
            "examples/with_typedef/interfaces.yang",
            "examples/with_typedef/expected.py",
            [],
            id="typedef",
        ),
        param(
            "examples/with_leafref/interfaces.yang",
            "examples/with_leafref/expected.py",
            [],
            id="leafref",
        ),
        param(
            "examples/with_restrictions/interfaces.yang",
            "examples/with_restrictions/expected.py",
            [],
            id="restrictions",
        ),
        param(
            "examples/with_uses/interfaces.yang",
            "examples/with_uses/expected.py",
            [],
            id="uses",
        ),
        param(
            "examples/with_case/interfaces.yang",
            "examples/with_case/expected.py",
            [],
            id="case",
        ),
        param(
            "examples/with_complex_case/interfaces.yang",
            "examples/with_complex_case/expected.py",
            [],
            id="complex case",
        ),
        param(
            "examples/turing-machine/turing-machine.yang",
            "examples/turing-machine/expected.py",
            [],
            id="turing machine",
        ),
        param(
            "examples/openconfig/openconfig-interfaces.yang",
            "examples/openconfig/expected.py",
            [
                "-t=openconfig-interfaces/interfaces/interface/config",
            ],
            id="openconfig",
        ),
        param(
            "examples/with_leaflist/interfaces.yang",
            "examples/with_leaflist/expected.py",
            [],
            id="leaf-list",
        ),
        param(
            "examples/with_union/interfaces.yang",
            "examples/with_union/expected.py",
            [],
            id="type union",
        ),
        param(
            "examples/with_identity_default/interfaces.yang",
            "examples/with_identity_default/expected.py",
            [],
            id="identity default",
        ),
        param(
            "examples/with_enum/interfaces.yang",
            "examples/with_enum/expected.py",
            [],
            id="enum",
        ),
    ],
)
def test_model(input_dir: str, expected_file: str, args: List[str], tmp_path: Path):
    input_folder = Path(__package__) / input_dir
    expected = Path(__package__) / expected_file
    run_pydantify(
        input_file=input_folder,
        output_folder=tmp_path,
        args=args,
    )
    print("Temp file: " + str(tmp_path / "out.py"))
    ParsedAST.assert_python_sources_equal(tmp_path / "out.py", expected)



from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field, RootModel
from typing_extensions import Annotated


class ComplexAddressLeaf(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
        regex_engine="python-re",
    )
    root: Annotated[
        str,
        Field(
            pattern="^(?=^(\\d{1,3}\\.){3}\\d{1,3}$).*$", title="Complex-addressLeaf"
        ),
    ]
    """
    Interface IP address. Example value: 10.10.10.1
    """


class ComplexPortLeaf(RootModel[int]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[int, Field(ge=1, le=65535, title="Complex-portLeaf")]
    """
    Port number. Example value: 8080
    """


class NameLeaf(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[str, Field(title="NameLeaf")]
    """
    Interface name. Example value: GigabitEthernet 0/0/0
    """


class SimplePortLeaf(RootModel[int]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[int, Field(ge=1, le=65535, title="Simple-portLeaf")]
    """
    Port number. Example value: 8080
    """


class InterfacesContainer(BaseModel):
    """
    Just a simple example of a container.
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    name: Annotated[NameLeaf, Field(alias="interfaces:name")]
    complex_address: Annotated[
        ComplexAddressLeaf, Field(alias="interfaces:complex-address")
    ]
    complex_port: Annotated[ComplexPortLeaf, Field(alias="interfaces:complex-port")]
    simple_port: Annotated[SimplePortLeaf, Field(alias="interfaces:simple-port")]


class Model(BaseModel):
    """
    Initialize an instance of this class and serialize it to JSON; this results in a RESTCONF payload.

    ## Tips
    Initialization:
    - all values have to be set via keyword arguments
    - if a class contains only a `root` field, it can be initialized as follows:
        - `member=MyNode(root=<value>)`
        - `member=<value>`

    Serialziation:
    - `exclude_defaults=True` omits fields set to their default value (recommended)
    - `by_alias=True` ensures qualified names are used (necessary)
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    interfaces: Annotated[
        InterfacesContainer, Field(None, alias="interfaces:interfaces")
    ]


if __name__ == "__main__":
    model = Model(  # type: ignore[call-arg]
        # <Initialize model here>
    )

    restconf_payload = model.model_dump_json(
        exclude_defaults=True, by_alias=True, indent=2
    )

    print(f"Generated output: {restconf_payload}")

    # Send config to network device:
    # from pydantify.utility import restconf_patch_request
    # restconf_patch_request(url='...', user_pw_auth=('usr', 'pw'), data=restconf_payload)



from enum import Enum
from typing import Dict, List, Type, Optional, Union

from pyang.statements import Statement, TypedefStatement, TypeStatement
from pyang.types import (
    BooleanTypeSpec,
    EmptyTypeSpec,
    EnumTypeSpec,
    IntTypeSpec,
    LengthTypeSpec,
    PathTypeSpec,
    PatternTypeSpec,
    RangeTypeSpec,
    StringTypeSpec,
    TypeSpec,
    XSDPattern,
    # TODO: Implement the following:
    BinaryTypeSpec,
    BitsTypeSpec,
    BitTypeSpec,
    EnumerationTypeSpec,
    IdentityrefTypeSpec,
    InstanceIdentifierTypeSpec,
    LeafrefTypeSpec,
    UnionTypeSpec,
    Decimal64TypeSpec,
)
from pydantic import Field
from pydantic.types import conbytes, confloat, conint, constr
from pydantic_core import PydanticUndefinedType, core_schema
from typing_extensions import Annotated, Self

from . import Node
from ..utility.patterns import convert_pattern


class TypeResolver:
    __mapping: Dict[Statement, Node] = dict()

    @classmethod
    def get_model_if_known(cls: Type[Self], stm: Statement) -> Node | None:
        return TypeResolver.__mapping.get(stm, None)

    @classmethod
    def register(cls: Type[Self], stm: Statement, model: Node):
        assert isinstance(model, Node) and isinstance(stm, Statement)
        cls.__mapping[stm] = model

    @classmethod
    def resolve_statement(cls: Type[Self], stm: Statement) -> type | Node | Enum:
        # Check if already known
        ret: Optional[Node] = cls.__mapping.get(stm, None)
        if ret is not None:
            return ret

        # If not known, check type definition
        stm_type: TypeStatement = stm.search_one(keyword="type")

        # # If we need to deal with identity in a specific way:
        # if stm_type is None:
        #     if stm.keyword == "identity":
        #         return object
        #     raise NotImplementedError(f"Statement {stm} not implemented")

        return cls.__resolve_type_statement(stm_type=stm_type)

    @classmethod
    def __resolve_type_statement(
        cls: Type[Self], stm_type: TypeStatement
    ) -> type | Enum:
        typespec: TypeSpec = getattr(stm_type, "i_type_spec", None)
        typedef: TypedefStatement = getattr(stm_type, "i_typedef", None)

        if typedef is not None:  # Type is a typedef
            ret = cls.__mapping.get(typedef, None)
            if ret is None:
                from . import TypeDefNode

                typedef_node: Node = TypeDefNode(typedef)
                cls.register(typedef, typedef_node)
                output_type = typedef_node._output_model.cls
            else:
                output_type = ret._output_model.cls
            if isinstance(output_type, PydanticUndefinedType):
                raise Exception(f"{typedef_node} output model is Undefined")
            return output_type

        if typespec is not None:  # Type is a base type
            resolved = cls.__resolve_type_spec(typespec)
            return resolved

        assert False  ## Not yet implemented

    @classmethod
    def __resolve_type_spec(cls: Type[Self], spec: TypeSpec) -> type | Enum:
        from . import Node, NodeFactory

        match (spec.__class__.__qualname__):
            case RangeTypeSpec.__qualname__:
                return conint(ge=spec.min, le=spec.max)
            case LengthTypeSpec.__qualname__:
                return constr(
                    min_length=spec.min,
                    max_length=spec.max,
                )
            case EnumTypeSpec.__qualname__:
                return Enum(
                    Node.ensure_unique_name(f"{spec.name}Enum"),
                    {x: x for x, _ in spec.enums},
                )  # TODO: make separate node type
            case PathTypeSpec.__qualname__:
                target_statement = getattr(spec, "i_target_node")
                if cls.__mapping.get(target_statement, None) is None:
                    NodeFactory.generate(target_statement)
                node = cls.__mapping.get(target_statement)
                if isinstance(node, Node):
                    return node._output_model.cls  # type: ignore
            case IntTypeSpec.__qualname__:
                return conint(
                    ge=spec.min,
                    le=spec.max,
                )
            case Decimal64TypeSpec.__qualname__:
                return confloat(
                    ge=spec.min.value,
                    le=spec.max.value,
                )
            case StringTypeSpec.__qualname__:
                return str
            case BooleanTypeSpec.__qualname__:
                return bool
            case BinaryTypeSpec.__qualname__:
                return conbytes(
                    min_length=spec.min,
                    max_length=spec.max,
                )
            case PatternTypeSpec.__qualname__:
                pattern = cls.__resolve_pattern(patterns=spec.res)
                return constr(pattern=convert_pattern(pattern))
            case EmptyTypeSpec.__qualname__:
                return dict
            case (
                IdentityrefTypeSpec.__qualname__
            ):  # TODO: abort before entering this stage?
                # def find_identityref(spec: IdentityrefTypeSpec):
                #     from pyang.util import prefix_to_module

                #     base_statement: BaseStatement = spec.idbases[0]
                #     base_arg: str = spec.idbases[0].arg
                #     module = base_statement.i_module
                #     pos = base_statement.pos
                #     errors = []

                #     prefix = None
                #     if ":" in base_arg:
                #         prefix, name = base_arg.split(":")
                #     else:
                #         name = base_arg

                #     if not prefix or prefix == module.i_module:
                #         prefix_module = module
                #     else:
                #         prefix_module = prefix_to_module(module, prefix, pos, errors)
                #         if prefix_module is None:
                #             raise Exception(f"No module found for prefix {prefix}")
                #     if stm := prefix_module.i_identities.get(name):
                #         return cls.resolve_statement(stm)
                #     raise Exception(f"No node {name} not found with prefix {prefix}")

                # return find_identityref(spec)
                return object
            case UnionTypeSpec.__qualname__:
                union = tuple([cls.__resolve_type_statement(typ) for typ in spec.types])
                return Union[union]  # type: ignore
        assert False, f'Spec "{spec.__class__.__qualname__}" not yet implemented.'

    @classmethod
    def __resolve_pattern(cls, patterns: List[XSDPattern]) -> str:
        comnbined_pattern: str = "^"
        for pattern in patterns:
            pattern_spec: str = pattern.spec
            if not pattern_spec.startswith("^"):
                pattern_spec = "^" + pattern_spec
            if not pattern_spec.endswith("$"):
                pattern_spec += "$"
            comnbined_pattern += f"(?={pattern_spec})"
        return comnbined_pattern + ".*$"  # Capture everything if all lookaheads suceed



import inspect
from typing import Callable
import re


def function_to_source_code(f: Callable) -> str:
    return inspect.getsource(f)


def function_content_to_source_code(f: Callable) -> str:
    function_lines: list[str] = inspect.getsourcelines(f)[0]
    function_content = function_lines[1:]
    indentation = re.compile("^([\t ]+)").findall(function_content[0])[0]
    return "".join(line.replace(indentation, "", 1) for line in function_content)



#!/usr/bin/env python

from typing import List
import logging
import sys


logging.basicConfig(
    stream=sys.stdout,
    format="[%(levelname)s] %(pathname)s:%(lineno)d (%(funcName)s): %(message)s",
    level=logging.INFO,
)
logger = logging.getLogger("pydantify")


def main():
    from pyang.scripts.pyang_tool import run

    # Parse user-give settings
    sys.argv[1:] = parse_cli_arguments()

    run()


def parse_cli_arguments() -> List[str]:
    """Parses and handles incoming CLI arguments and returns all the ones destined for Pydantic."""

    # Import locally to not clutter scope of caller
    import logging
    from argparse import ArgumentParser
    from .utility.model_generator import ModelGenerator
    from pathlib import Path
    import os

    # Setup parser
    parser = ArgumentParser(
        prog="pydantify",
        description="Transform a YANG model to a serializable Pydantic model.",
        epilog="NOTE: All unknown arguments will be passed to Pyang as-is and without guarantees.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        help="Enables debug output",
        default=False,
    )
    parser.add_argument(
        "-V",
        "--include-verification",
        action="store_true",
        dest="verify",
        help="Adds validation code, as well as the relevant YANG files, to the output model.",
        default=False,
    )
    parser.add_argument(
        "-S",
        "--standalone",
        action="store_true",
        dest="standalone",
        help="Generated output model has no dependency on Pydantify. All required code is copied into the output model.",
        default=False,
    )
    parser.add_argument(
        "-i",
        "--input-dir",
        "--path",
        dest="input_dir",
        help="The directory that contains the YANG input model. Defaults to the input file's folder.",
        default=None,
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="output_dir",
        help='The directory that should be used to store the output model. Defaults to "$CWD/out".',
        default=f"{os.getcwd()}/out/",
    )
    parser.add_argument(
        "input_file",
        action="store",
        help="The YANG file containing the entrypoint to the model to evaluate.",
    )
    parser.add_argument(
        "-t",
        "--trim-path",
        dest="trim_path",
        help="Get only the specified branch of the whole tree.",
        default=None,
    )
    relay_args: List[str] = []

    # Parse
    args, unknown_args = parser.parse_known_args()

    # Apply known settings accordingly
    logger.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    ModelGenerator.include_verification_code = args.verify
    ModelGenerator.standalone = args.standalone

    input_dir = (
        Path(args.input_file).absolute().parent
        if args.input_dir is None
        else Path(args.input_dir).absolute()
    )
    ModelGenerator.input_dir = input_dir
    relay_args.append(f"--path={input_dir}")

    ModelGenerator.trim_path = args.trim_path

    output_dir = Path(args.output_dir).absolute()
    ModelGenerator.output_dir = output_dir
    os.makedirs(output_dir, exist_ok=True)  # Create output directory if not exists
    with open(output_dir / "__init__.py", "a"):  # Create init file if not exists
        pass
    relay_args.append(f"--output={output_dir}/out.py")

    relay_args.append(f"--plugindir={Path(__file__).parent}/plugins")
    relay_args.append("--format=pydantic")

    # Order of arguments matters.
    return relay_args + unknown_args + [args.input_file]



from __future__ import annotations

from enum import Enum
from typing import List

from pydantic import BaseModel, ConfigDict, Field, RootModel
from typing_extensions import Annotated


class CellIndexType(RootModel[int]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[int, Field(ge=-9223372036854775808, le=9223372036854775807)]
    """
    Type for indexing tape cells.
    """


class CoordLeaf(RootModel[CellIndexType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[CellIndexType, Field(title="CoordLeaf")]
    """
    Coordinate (index) of the tape cell.
    """


class HeadPositionLeaf(RootModel[CellIndexType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[CellIndexType, Field(title="Head-positionLeaf")]
    """
    Position of tape read/write head.
    """


class LabelLeaf(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[str, Field(title="LabelLeaf")]
    """
    An arbitrary label of the transition rule.
    """


class StateIndexType(RootModel[int]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[int, Field(ge=0, le=65535)]
    """
    Type for indexing states of the control unit.
    """


class StateLeaf(RootModel[StateIndexType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[StateIndexType, Field(title="StateLeaf")]
    """
    Current state of the control unit.
    The initial state is 0.
    """


class StateLeaf2(RootModel[StateIndexType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[StateIndexType, Field(title="StateLeaf2")]
    """
    Current state of the control unit.
    """


class StateLeaf3(RootModel[StateIndexType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[StateIndexType, Field(title="StateLeaf3")]
    """
    New state of the control unit. If this leaf is not
    present, the state doesn't change.
    """


class TapeSymbolType(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[str, Field(max_length=1, min_length=0)]
    """
    Type of symbols appearing in tape cells.
    A blank is represented as an empty string where necessary.
    """


class EnumerationEnum(Enum):
    integer_0 = 0
    integer_1 = 1


class HeadDirType(RootModel[EnumerationEnum]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: EnumerationEnum
    """
    Possible directions for moving the read/write head, one cell
    to the left or right (default).
    """


class HeadMoveLeaf(RootModel[HeadDirType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[HeadDirType, Field(title="Head-moveLeaf")]
    """
    Move the head one cell to the left or right
    """


class SymbolLeaf(RootModel[TapeSymbolType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[TapeSymbolType, Field(title="SymbolLeaf")]
    """
    Symbol appearing in the tape cell.
    Blank (empty string) is not allowed here because the
    'cell' list only contains non-blank cells.
    """


class SymbolLeaf2(RootModel[TapeSymbolType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[TapeSymbolType, Field(title="SymbolLeaf2")]
    """
    Symbol read from the tape cell.
    """


class SymbolLeaf3(RootModel[TapeSymbolType]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[TapeSymbolType, Field(title="SymbolLeaf3")]
    """
    Symbol to be written to the tape cell. If this leaf is
    not present, the symbol doesn't change.
    """


class CellListEntry(BaseModel):
    """
    List of non-blank cells.
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    coord: Annotated[CoordLeaf, Field(None, alias="turing-machine:coord")]
    symbol: Annotated[SymbolLeaf, Field(None, alias="turing-machine:symbol")]


class InputContainer(BaseModel):
    """
    Input parameters (arguments) of the transition rule.
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    state: Annotated[StateLeaf2, Field(alias="turing-machine:state")]
    symbol: Annotated[SymbolLeaf2, Field(alias="turing-machine:symbol")]


class OutputContainer(BaseModel):
    """
    Output values of the transition rule.
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    state: Annotated[StateLeaf3, Field(None, alias="turing-machine:state")]
    symbol: Annotated[SymbolLeaf3, Field(None, alias="turing-machine:symbol")]
    head_move: Annotated[HeadMoveLeaf, Field("right", alias="turing-machine:head-move")]


class TapeContainer(BaseModel):
    """
    The contents of the tape.
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    cell: Annotated[List[CellListEntry], Field(alias="turing-machine:cell")]


class DeltaListEntry(BaseModel):
    """
    The list of transition rules.
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    label: Annotated[LabelLeaf, Field(None, alias="turing-machine:label")]
    input: Annotated[InputContainer, Field(None, alias="turing-machine:input")]
    output: Annotated[OutputContainer, Field(None, alias="turing-machine:output")]


class TransitionFunctionContainer(BaseModel):
    """
    The Turing Machine is configured by specifying the
    transition function.
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    delta: Annotated[List[DeltaListEntry], Field(alias="turing-machine:delta")]


class TuringMachineContainer(BaseModel):
    """
    State data and configuration of a Turing Machine.
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    state: Annotated[StateLeaf, Field(alias="turing-machine:state")]
    head_position: Annotated[
        HeadPositionLeaf, Field(alias="turing-machine:head-position")
    ]
    tape: Annotated[TapeContainer, Field(None, alias="turing-machine:tape")]
    transition_function: Annotated[
        TransitionFunctionContainer,
        Field(None, alias="turing-machine:transition-function"),
    ]


class Model(BaseModel):
    """
    Initialize an instance of this class and serialize it to JSON; this results in a RESTCONF payload.

    ## Tips
    Initialization:
    - all values have to be set via keyword arguments
    - if a class contains only a `root` field, it can be initialized as follows:
        - `member=MyNode(root=<value>)`
        - `member=<value>`

    Serialziation:
    - `exclude_defaults=True` omits fields set to their default value (recommended)
    - `by_alias=True` ensures qualified names are used (necessary)
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    turing_machine: Annotated[
        TuringMachineContainer, Field(None, alias="turing-machine:turing-machine")
    ]


if __name__ == "__main__":
    model = Model(  # type: ignore[call-arg]
        # <Initialize model here>
    )

    restconf_payload = model.model_dump_json(
        exclude_defaults=True, by_alias=True, indent=2
    )

    print(f"Generated output: {restconf_payload}")

    # Send config to network device:
    # from pydantify.utility import restconf_patch_request
    # restconf_patch_request(url='...', user_pw_auth=('usr', 'pw'), data=restconf_payload)



from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field, RootModel
from typing_extensions import Annotated


class AddressLeaf(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[str, Field(title="AddressLeaf")]
    """
    Interface IP address. Example value: 10.10.10.1
    """


class NameLeaf(RootModel[str]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[str, Field(title="NameLeaf")]
    """
    Interface name. Example value: GigabitEthernet 0/0/0
    """


class PortLeaf(RootModel[int]):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    root: Annotated[int, Field(ge=0, le=65535, title="PortLeaf")]
    """
    Port number. Example value: 8080
    """


class InterfacesContainer(BaseModel):
    """
    Just a simple example of a container.
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    name: Annotated[NameLeaf, Field(alias="interfaces:name")]
    address: Annotated[AddressLeaf, Field(alias="interfaces:address")]
    port: Annotated[PortLeaf, Field(alias="interfaces:port")]


class Model(BaseModel):
    """
    Initialize an instance of this class and serialize it to JSON; this results in a RESTCONF payload.

    ## Tips
    Initialization:
    - all values have to be set via keyword arguments
    - if a class contains only a `root` field, it can be initialized as follows:
        - `member=MyNode(root=<value>)`
        - `member=<value>`

    Serialziation:
    - `exclude_defaults=True` omits fields set to their default value (recommended)
    - `by_alias=True` ensures qualified names are used (necessary)
    """

    model_config = ConfigDict(
        populate_by_name=True,
    )
    interfaces: Annotated[
        InterfacesContainer, Field(None, alias="interfaces:interfaces")
    ]


def restconf_patch_request(url: str, user_pw_auth: tuple[str, str], data: str):
    r"""Sends a restconf "PATCH" request to a network device

    :param url: e.g. `'https://sandbox.cisco.com/restconf/data/Cisco-IOS-XE-native:native/interface'`
    :param user_pw_auth: tuple of username & password, e.g. `("user", "pw")`
    :param data: restconf payload as generated by `Model.json(...)`

    :return: HTTP Request response
    """
    import requests

    response = requests.patch(
        url=url,
        auth=user_pw_auth,
        headers={
            "Accept": "application/yang-data+json",
            "Content-Type": "application/yang-data+json",
        },
        data=data,
        verify=False,
    )

    return response


if __name__ == "__main__":
    model = Model(  # type: ignore[call-arg]
        # <Initialize model here>
    )

    restconf_payload = model.model_dump_json(
        exclude_defaults=True, by_alias=True, indent=2
    )

    print(f"Generated output: {restconf_payload}")

    # Send config to network device:
    # from pydantify.utility import restconf_patch_request
    # restconf_patch_request(url='...', user_pw_auth=('usr', 'pw'), data=restconf_payload)



from .base import BaseModel, GeneratedClass, Node, RootModel
from .typeresolver import TypeResolver
from .nodefactory import NodeFactory
from .models import (
    ContainerNode,
    GeneratedClass,
    LeafNode,
    ListNode,
    ModelRoot,
    ModuleNode,
    TypeDefNode,
)
